import evaluate
import torch
from statistics import mean
import torch
from tqdm import tqdm
from transformers import AutoTokenizer, AutoModelForSequenceClassification, AdamW, get_scheduler
import argparse
import pickle as pkl
import random
import os
import numpy as np
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score
#import accelerate
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler, Dataset

device=torch.device("cuda" if torch.cuda.is_available() else "cpu")
#from accelerate import Accelerator

#accelerator=Accelerator()
parser=argparse.ArgumentParser()



def main():
	tokenizer=AutoTokenizer.from_pretrained(args.tokenizer)
	#model=torch.load(args.model)
	#print(type(model))
	model=AutoModelForSequenceClassification.from_pretrained(args.model)
	
	with open(args.dataset, "rb") as f:
		data=pkl.load(f)
	print(len(data))	
	dataset=AKIDataset(data, tokenizer)
	evaluate_model_loop(model, args.seed, dataset, args.batch_size, args.metrics)


class AKIDataset(torch.utils.data.Dataset):
    def __init__(self, instances, tokenizer):
        self.instances=instances
        self.tokenizer=tokenizer
    def __getitem__(self, idx):
        instance=self.instances[idx]
        #print(instance)
        prompt=instance[0]
        labels=instance[1]
        item=self.tokenize(prompt)
        item["label"] = labels
        #print(item["input_ids"].size())
        #print(item["input_ids"].size())
        #item["input_ids"]
        #item["label"]#.to(device)
        #item["attention_mask"]
        return item

    def tokenize(self, prompt):
        result=self.tokenizer(prompt, truncation=True, max_length=512, padding="max_length", return_tensors="pt")
        return result
    def __len__(self):
        return len(self.instances)

def set_seed(seed):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic=True
    torch.backends.cudnn.banchmark=False
    os.environ["PYTHONHASHSEED"]=str(seed)
    print("Random seed set as : {seed}")




def evaluate_model_loop(model, seed, test_dataset, batch_size, metrics):
	torch.cuda.empty_cache()
	set_seed(seed)
	print("Testing model")
	print("Metric: ", metrics)
	#if len(metrics) > 1:
	#	metric=evaluate.combine(metrics)
	#else:
	#	print("here")
	#	metric=evaluate.load(metrics[0])
	#metric=evaluate.combine([evaluate.load("accuracy", average=None), 
	#						 evaluate.load("f1",average=None),
	#						 evaluate.load("precision", average=None),
	#						 evaluate.load("recall", average=None)])
	
	metric=evaluate.load("accuracy", average=None)
	model.eval().to(device)
	print("Batch size: ", batch_size)
	print("metrics: ", metrics)
	all_accs=[]
	all_precs=[]
	all_recs=[]
	all_f1=[]

	#config=GenerationConfig.from_pretrained(config_name, top_k=top_k, temperature=temp, num_beams=num_beams, early_stopping=early_stopping, no_repeat_ngram_size=no_rep, num_return_sequences=num_return_sequences, max_length=512)
	#print(config.num_return_sequences)
	#print(tokenizer.max_length)
	eval_sampler=SequentialSampler(test_dataset)
	eval_dataloader=DataLoader(test_dataset, sampler=eval_sampler, batch_size=batch_size)
	iterations=0
	for index, batch in tqdm(enumerate(eval_dataloader)):
		with torch.no_grad():
			#print(batch.to(device))
			#if index == 10:
			#	break
			input_ids=batch["input_ids"].squeeze().to(device)
			#print(input_ids.size())
			attention_mask=batch["attention_mask"].squeeze().to(device)
			token_type_ids=batch["token_type_ids"].squeeze().to(device)
			labels=batch["label"].to(device)

			outputs=model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
			prediction=torch.argmax(outputs.logits.softmax(1), dim=-1)
			print(prediction)
			print(outputs.logits.softmax(1))
			accuracy_res=metric.compute(predictions=prediction, references=labels)
			print(accuracy_res)
			all_accs.append(accuracy_res["accuracy"])
			iterations+=1
			#all_accs.append(accuracy_score(labels.cpu().squeeze(), prediction.cpu().squeeze()))
			#all_precs.append(precision_score(labels.cpu().squeeze(), prediction.cpu().squeeze()))
			#all_recs.append(recall_score(labels.cpu().squeeze(), prediction.cpu().squeeze()))
			#all_f1.append(f1_scorer(labels.cpu().squeeze(), prediction.cpu().squeeze()))
	#print(len(eval_dataloader))
	#print(len(list(eval_dataloader.items())))
	#print(sum(all_accs))
	acc=sum(all_accs)/iterations
	#prec=sum(all_precs)/len(eval_dataloader)
	#rec=sum(all_recs)/len(eval_dataloader)
	#f1=sum(all_f1)/len(eval_dataloader)	

	print(acc)
	#res=metric.compute(average=None)
	#print("RESULTS:\n\nAccuracy: {0}\nPrecision: {1}\nRecall: {2}\nF1: {3}".format(round(acc*100, 2),round(prec*100, 2), round(rec*100, 2), round(f1*100, 2)))
	return None


if __name__ == "__main__":
	parser.add_argument("--model", type=str)
	parser.add_argument("--tokenizer", type=str)
	parser.add_argument("--dataset", type=str)
	parser.add_argument("--seed", type=int)
	parser.add_argument("--batch_size", type=int)
	parser.add_argument("--metrics", nargs="+")

	args=parser.parse_args()
	main()

