from transformers import pipeline, AutoTokenizer, AutoModelForCausalLM
import argparse
import pickle as pkl
import accelerate
import evaluate
from statistics import mean
from tqdm import tqdm
import torch
import re
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler

device=torch.device("cuda" if torch.cuda.is_available() else "cpu")
def main():
    #first, we need to load the dataset and instantiate the pipeline for inference
    pipe=pipeline(model=args.model,model_kwargs={"device_map":"auto", "torch_dtype":torch.float16},  max_new_tokens=3)
    tokenizer=AutoTokenizer.from_pretrained(args.tokenizer)
    #with open(args.input_dataset, "rb") as f:
     #   input_dataset=pkl.load(f)
    
    with open(args.dataset, "rb") as t:
        test_dataset=pkl.load(t)
    test_dataset=InferenceDataset(test_dataset, tokenizer)
    
    model=AutoModelForCausalLM.from_pretrained(args.model, device_map="auto", torch_dtype=torch.float16)
    model.tie_weights()
    #Now, we need to call the function that will generate and evaluate the outputs of the mode
    print("Classidy in main: ", args.classify)
    evaluate_inference(model, tokenizer, test_dataset, args.top_k, args.num_beams, args.temp, args.num_return_sequences, args.early_stopping, classify=args.classify)


def evaluate_inference(model,tokenizer, test_dataset, top_k, num_beams, temp, num_return_sequences, early_stopping, classify=False, batch_size=1):
    print("classify? ", classify)
    #for dp in tqdm(test_dataset):
        #if classify == False:
    #for output in pipeline_model(test_dataset, top_k=top_k, num_beams=num_beams, temperature=temp, num_return_sequences=num_return_sequences, early_stopping=early_stopping):
     #   print(output)
    
        #elif classify == True:
         #   output=pipeline_model(dp["prompt"], candiate_labels=["1", "2", "3"])

        #print(output)

    #We will use generate function instead and evaluate as normal

    torch.cuda.empty_cache()
    eval_sampler=SequentialSampler(test_dataset)
    eval_dataloader=DataLoader(test_dataset, sampler=eval_sampler, batch_size=batch_size)
    all_accs=[]
    iterations=0
    recalls=[]
    precisions=[]
    f1_micro=[]
    f1_macro=[]
    metric=evaluate.load("accuracy", average=None)
    #metric_f1_micro=evaluate.load("f1", average="micro")
    #metric_f1_macro=evaluate.load("f1", average="macro")
   
    for index, batch in tqdm(enumerate(eval_dataloader)):
        with torch.no_grad():
            input_ids=torch.tensor(batch["input_ids"]).unsqueeze(0).to(device)
            attention_mask=torch.tensor(batch["attention_mask"]).unsqueeze(0).to(device)
            labels=torch.tensor(batch["labels"]).unsqueeze(0).to(device)

            #print(input_ids.shape)
            #print(attention_mask.size)
            #print(batch["labels"].size())
            labels_len=len([el for el in batch["labels"] if el != -100])
            input_ids_len=len(batch["input_ids"][:-1])
            #print(input_ids_len)
            #print(labels_len)



            outputs=model.generate(input_ids=input_ids, attention_mask=attention_mask, top_k=top_k, temperature=temp, num_beams=num_beams, early_stopping=early_stopping, no_repeat_ngram_size=1, num_return_sequences=num_return_sequences, min_length=input_ids_len + 2, max_new_tokens=8) #labels_len)
            preds=tokenizer.decode(outputs[0][input_ids_len:], skip_special_tokens=True)
            #print("Predictions: ", preds)
            print("Whole output: ", tokenizer.decode(outputs[0], skip_special_tokens=True))
            #print("Label: ", tokenizer.decode([el.item() for el in batch["labels"] if el != -100]))
            clean_label=tokenizer.decode([el.item() for el in batch["labels"] if el != -100])


            match=re.search(r"\d", preds)
            if classify == True:
                 if match:
                    cleaned_pred=match.group()
                 else:
                     print("No digit")
                 accuracy_res=metric.compute(predictions=cleaned_pred, references=clean_label)
                 #f1_macro_res=metric.compute(predictions=cleaned_pred, references=clean_label)
                
                 all_accs.append(accuracy_res["accuracy"])
                 iterations+=1
            else:
                preds_list=preds.split(", ")
                labels_list=clean_label.split(", ")

                preds_clean=[p.strip(" ").lstrip(" ").lower() for p in preds_list]
                labels_clean=[l.strip(" ").lstrip(" ").lower() for l in labels_list]

                print("Predictions: ", preds_clean)
                print("Labels: ", labels_clean)
                TP=len([t for t in preds_clean if t in labels_clean])
                FP=len([f for f in preds_clean if f not in labels_clean])
                FN=len([fn for fn in labels_clean if fn not in preds_clean])

                p=TP/(TP+FP) if (TP+FP) != 0 else 0
                r=TP/(TP+FN) if (TP+FN) != 0 else 0

                precisions.append(p)
                recalls.append(r)

            #accuracy_res=metric.compute(predictions=cleaned_pred, references=clean_label)
               
           
            #all_accs.append(accuracy_res["accuracy"])
            #iterations+=1
    if classify== True:
 
        acc=sum(all_accs)/iterations
        #prec=mean(precisions)
        #rec=mean(recalls)
        #f1=(2*prec*rec)/(prec+rec)
        #print("F1: ", f1)
        #print("Recall: ", rec)
        #print("Precision: ", prec)
        print("Accuracy: ", acc)
    elif classify==False:
        prec=mean(precisions)
        rec=mean(recalls)
        f1=(2*prec*rec)/(prec+rec)
        print("RESULTS: Precision: " + str(prec) + "Recall: " + str(rec) + "F1: " + str(f1))
    return None
class InferenceDataset(torch.utils.data.Dataset):
    def __init__(self, instances, tokenizer):
        self.instances=instances
        self.tokenizer=tokenizer
    def __getitem__(self, idx):
        instance=self.instances[idx] #get element
        prompt=instance["prompt"]
        labels=instance["label"]
        item=self.tokenize(prompt + labels)
        tokenized_instruction=self.tokenize(prompt)
        tokenized_labels=self.tokenize(labels)


        item2=dict()
        item2["labels"]=tokenized_labels["input_ids"]
        item2["input_ids"]=tokenized_instruction["input_ids"]
        item2["attention_mask"]=tokenized_instruction["attention_mask"]


        return item2
    
    def tokenize(self, prompt):
        result_prompts=self.tokenizer(prompt, truncation=True, max_length=2048, padding=False, return_tensors=None)
        result_prompts["labels"]=[-100]*len(result_prompts["input_ids"])
        return result_prompts
    def __len__(self):
        return len(self.instances)


if __name__ == "__main__":
    parser=argparse.ArgumentParser()
    parser.add_argument("--model", type=str)
    parser.add_argument("--tokenizer", type=str)
    #parser.add_argument("--input_dataset", type=str)
    parser.add_argument("--dataset", type=str)
    parser.add_argument("--top_k", type=float, default=30)
    parser.add_argument("--temp", type=float, default=0.9)
    parser.add_argument("--num_beams", type=int, default=4)
    parser.add_argument("--num_return_sequences", type=int, default=1)
    parser.add_argument("--early_stopping", action="store_true")
    parser.add_argument("--classify", action="store_true") 
    parser.add_argument("--metrics", nargs="+")
    args=parser.parse_args()
    main()
