"""T5 Finetuning Model for discharge summary"""
from tqdm import tqdm
from transformers import T5ForConditionalGeneration, T5Tokenizer, AutoTokenizer, AutoModelForCausalLM, AutoConfig, T5Config, Trainer, DataCollatorForSeq2Seq, GenerationConfig
import random
from datasets import Dataset
import os
from transformers import BitsAndBytesConfig
import numpy as np
import pandas as pd
#from pandas import DataFrameDataset
import torch
from statistics import mean
import pickle as pkl
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler, TensorDataset, IterableDataset
import logging
from transformers import get_scheduler, AdamW, TrainingArguments, Seq2SeqTrainer, Seq2SeqTrainingArguments
from torch.nn.parallel import DataParallel 
#import torch.distributed as dist
#import torch.multiprocessing as mp
import argparse
from trl import SFTTrainer
import trl
from trl.trainer import ConstantLengthDataset
from peft import get_peft_config, get_peft_model, get_peft_model_state_dict, LoraConfig, TaskType, prepare_model_for_int8_training
#from evaluate_model import evaluate_model_loop
#from torch.nn import CrossEntropyLoss
from accelerate import infer_auto_device_map, Accelerator, init_empty_weights
#logging.basicConfig(filename='log_t5.log',  level=logging.DEBUG)
#device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
#available_gpus=torch.cuda.device_count()
accelerator = Accelerator()
device =accelerator.device
#print(device.index)
#offload="./"
#def main():
 #   num_gpus=2
  #      mp.spawn(run, args=(2,), nprocs=2, join=True) 
tokenizer=AutoTokenizer.from_pretrained("medalpaca/medalpaca-7b")
def run():
    logging.basicConfig(filename=args.log, level=logging.DEBUG)
    logging.info("Starting Training")
    print("Started Training")
    bnb_config=BitsAndBytesConfig(load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype="float16",
                bnb_4bit_use_double_quant=False,
                )

    print(trl.__version__)
    print(device)


    tokenizer=AutoTokenizer.from_pretrained(args.tokenizer) #, torch_dtype=torch.float16, device_map="auto") #, trust_remote_code=True)
    #tokenizer.pad_token=tokenizer.eos_token
    #print(tokenizer.pad_token)
    #print(tokenizer.pad_token_id)
    #print(tokenizer.eos_token_id)
    #tokenizer.add_tokens(["<A>", "<G>", "<M>", "<T>", "<L>", "<P>", "<E>"], special_tokens=True)
    #config=AutoConfig.from_pretrained("facebook/opt-13b")
    #with init_empty_weights():
    #	model=AutoModelForCausalLM.from_config(config)
    #device_map=infer_auto_device_map(model, no_split_module_classes=["OPTDecoderLayer"])
    #device_map["lm_head"]=device_map["decoder.embed_tokens"]

    #print(device_map)
    model=AutoModelForCausalLM.from_pretrained(args.model, torch_dtype=torch.float16, device_map="auto") #, quantization_config=bnb_config)#, trust_remote_code=True)
    #model=prepare_model_for_int8_training(model)
    #model=T5ForConditionalGeneration.from_pretrained(args.model, from_flax=args.ff)
    #peft_config=LoraConfig(task_type="SEQ_2_SEQ", inference_mode=False, r=4, lora_alpha=16, lora_dropout=0.1, bias="none", target_modules=["query_key_value"])
    #model=get_peft_model(model, peft_config)
    #model.print_trainable_parameters()

    training_arguments=Seq2SeqTrainingArguments(
                output_dir="./medalpaca_output_dir",
                per_device_train_batch_size=1,
                per_device_eval_batch_size=1,
                gradient_accumulation_steps=2,
                optim="adamw_torch",
                save_strategy="epoch",
                evaluation_strategy="epoch",
                logging_steps=10,
                learning_rate=2e-4,
                num_train_epochs=args.epochs,
                fp16=True,
                bf16=False,
                predict_with_generate=True,
                generation_max_length=600,
                lr_scheduler_type="linear",
                warmup_ratio=0.1)
    peft_config=LoraConfig(
                lora_alpha=16,
                lora_dropout=0.05,
                r=8,
                bias="none",
                task_type=TaskType.CAUSAL_LM,
                target_modules=["q_proj", "v_proj"])
    #target_modules=["query_key_value"])

    #model.config.use_cache=False
    model=get_peft_model(model, peft_config)
    model.print_trainable_parameters()

    #model=DataParallel(model)


    #model=accelerator.prepare(model)
    #print(model)


    #model=BioGptForCausalLM.from_pretrained(args.model, from_flax=args.ff, device_map="auto")
    #torch.distributed.init_process_group(backend="nccl", rank=rank, world_size=world_size)
    #dist.get_rank()
    #print("Rank: ", dist.get_rank())
    ##device_id=rank% torch.cuda.device_count()
    #print(device_id)
    #model = torch.nn.DataParallel(model)
    #tokenizer=AutoTokenizer.from_pretrained(args.tokenizer, trust_remote_code=True)
    #tokenizer.pad_token=tokenizer.eos_token
    #print(tokenizer.pad_token)
    #tokenizer.add_tokens(["<A>", "<G>", "<M>", "<T>", "<L>", "<P>", "<E>"], special_tokens=True)
    #print(tokenizer.d_model)
    print("Model loaded")
    logging.info("Loading dataset")
    print(args.dataset)
    #dataset=preprocess(tokenizer, args.dataset)
    #dataset=pd.read_csv(args.dataset, usecols=["prompt", "label"])
    #print(dataset.head())
    with open(args.dataset, "rb") as f:
        dataset=pkl.load(f)
    with open(args.test_dataset, "rb") as t:
        test_dataset=pkl.load(t)
    test_df=pd.DataFrame(test_dataset)
    #dataset_dict=dataset.to_dict()
    df=pd.DataFrame(dataset)
    dataset_dict=df.to_dict(orient="list")
    print(df.head())
    test_dataset_dict=test_df.to_dict(orient="list")
    #pandas_dataset=Dataset.from_pandas(df)
    dataset=DiagnosesDataset(dataset, tokenizer, "train")
    test_dataset=DiagnosesDataset(test_dataset, tokenizer, "test")
    #print(pandas_dataset[0])
    #dataloader=DataLoader(pandas_dataset)
    #preprocessed_dataset=preprocess(tokenizer, args.dataset)	

    #prompts=[]
    #labels=[]
    #for el in dataset:
    #	prompts.append(el["prompt"])
    #	labels.append(el["label"])
    #dataset_class=DiagnosesDataset(prompts, labels)
    #print(dataset_class[0])
    #constant_len_dataset=ConstantLengthDataset(tokenizer, dataset_class,dataset_text_field="prompt", formatting_func=formatting_prompts_func)
    #print("okay")
    #constant_len_dataset=ConstantLengthDataset(dataset)

    trainer=Seq2SeqTrainer(
                model=model,
                train_dataset=dataset,
                eval_dataset=test_dataset,
                args=training_arguments,
                compute_metrics=compute_metrics_func,
                data_collator=DataCollatorForSeq2Seq(tokenizer, return_tensors="pt")) #save_strategy="epoch"
    model.use_cache=False
    trainer.train()

    #trainer=SFTTrainer(
    #	model=model,
    #	tokenizer=tokenizer,
    #	train_dataset=dataset,
    #	peft_config=peft_config,
    #	packing=False,
    #	data_collator=DataCollatorForSeq2Seq(tokenizer, pad_to_multiple_of=8, return_tensors="pt", padding="max_length", max_length=2048),
    #	args=training_arguments)
    #model.use_cache=False
    #trainer.train()

    logging.info("Run Train loop")
    #model_updated=train(model, dataset, args.seed, args.batch_size, args.epochs, args.learning_rate,args.gradient_accumulation_steps, args.model_save_path)
    logging.info("Training Done!")
    if args.do_eval == True:
        logging.info("Running Evaluation")
        logging.info("Preparing Dataset")
        test_dataset=preprocess(tokenizer, args.test_dataset)
        print(args.metrics)
        res=evaluate_model_loop(model_updated, test_dataset, args.test_batch_size, tokenizer, args.topk, args.temp, args.num_beams, args.early_stopping, args.no_rep_ngram, args.num_return_sequences, metrics=args.metrics, generative=True)
        logging.info(res)

class DataFrameDataset(IterableDataset):
    def __init__(self, dataframe):
        self.dataframe=dataframe
        def __iter__(self):
            for _, row in self.dataframe.iterrows():
                yield row

def compute_metrics_func(eval_preds):
    #print("here")
    preds, labels=eval_preds
    #print("Raw preds: ", preds)
    #print(labels.shape)
    #print(preds.shape)
    preds[preds == -100]=tokenizer.pad_token_id
    labels[labels==-100]=tokenizer.pad_token_id

    
    squeezed_preds=np.squeeze(preds).tolist()

    squeezed_labels=np.squeeze(labels).tolist()
    #print(squeezed_preds)
    #print(squeezed_labels)
    precs=[]
    recs=[]
    

    for i in range(len(squeezed_preds)):
        input_len=len([l for l in squeezed_labels[i] if l == tokenizer.pad_token_id])
        cropped_preds=squeezed_preds[i][input_len:]
        cropped_labels=[t for t in squeezed_labels[i] if t !=tokenizer.pad_token_id]
        #print("cropped preds: ", cropped_preds)
        #print("cropped labels: ", cropped_labels)
        decoded_preds=tokenizer.decode(cropped_preds, skip_special_tokens=True)
        decoded_labels=tokenizer.decode(cropped_labels, skip_special_tokes=True)
        #print("Decoded preds: ", decoded_preds)
        #print("Decoded labels: ", decoded_labels)

        decoded_preds_list=decoded_preds.split(", ")
        decoded_labels_list=decoded_labels.split(", ")
        TP=len([t for t in list(set(decoded_preds_list)) if t in decoded_labels_list])
        FP=len([f for f in list(set(decoded_preds_list)) if f not in decoded_labels_list])
        FN=len([fn for fn in decoded_labels_list if fn not in list(set(decoded_preds_list))])

        p=(TP/(TP+FP)) if (TP+FP) != 0 else 0
        r=(TP/(TP+FN)) if (TP+FN) != 0 else 0

        precs.append(p)
        recs.append(r)

        #print("\n\n") 
    #preds_list=preds.tolist()
    #print(len(preds_list))
    #cropped_labels = labels[labels != -100].tolist()
    #print("Labels shape: ", labels[labels==-100].shape)
    #input_text_len=labels[labels==-100].shape[0]
    #print(input_text_len)
    #print(preds.shape)
    #only_labels_from_preds=preds[input_text_len:]
    #print(only_labels_from_preds)
    #cropped_preds=only_labels_from_preds[only_labels_from_preds != -100]
    #print(cropped_labels.shape)
    #print(type(labels))
    #print(tokenizer.pad_token_id)
    #preds=tokenizer.batch_decode(preds, skip_special_tokens=True)
    #labels=tokenizer.batch_decode(labels, skip_special_tokens=True)

    #print("Deoded predictions: ", preds)
    #print("Decoded labels: ", labels)

    prec=mean(precs)
    rec=mean(recs)
    f1=(2*prec*rec)/(prec+rec)
    result_dict={"precision":prec, "recall":rec, "F1":f1}
    print("RESULTS: Precision: " + str(prec) + " Recall: " + str(rec) + " F1: " + str(f1))
    return result_dict

class DiagnosesDataset(torch.utils.data.Dataset):
    def __init__(self, instances, tokenizer, state):
        self.instances=instances
        #self.labels=labels
        self.tokenizer=tokenizer
        self.state=state

    def __getitem__(self, idx):
        item={}
        prompt= self.instances[idx]["prompt"]
        labels = self.instances[idx]["label"]

        item=self.tokenize(prompt+labels)

        tokenized_instruction=self.tokenize(prompt)
        assert len(tokenized_instruction) < len(item["input_ids"])
        label_instruction=self.tokenize(labels)
        #print("Tokenized instruction input ids: ", len(tokenized_instruction["input_ids"]))
        #print("Labels input ids: ", len(label_instruction["input_ids"]))
        #print("Labels Tokenized_instruction: ", len(tokenized_instruction["labels"]))
        #print("Labels: ", len(label_instruction["labels"]))
        
        
        i=len(tokenized_instruction["input_ids"])-1 #remove 1 because of len and 1 because of EOS token
        #item["input_ids"].pop(i)
        #item["labels"].pop(i)
        #item["attention_mask"].pop(i)

        #while 1<len(item["input_ids"])
        #print("I index: ", i)
        #print("len to be filled: ",  len(item["labels"]) - i)
        #print("Len of tokenized instructions: ", len(tokenized_instruction["input_ids"]))
        #print(self.tokenizer.decode(tokenized_instruction["input_ids"]))

        #print("Len of item labels before ", len(label_instruction["input_ids"]))
        #while i<len(item["input_ids"]):
         #   for j in range(i):
          #      item["labels"][i] = label_instruction["input_ids"][j]
           #     i+=1
        
        #print(len(item["input_ids"][:i]) + len(label_instruction["input_ids"]))
        #print(self.tokenizer.decode(item["labels"]))
        #print(self.tokenizer.decode(item["input_ids"]))
        item2=dict()

        if self.state=="train":
            #item2["labels"]=item["input_ids"][1:]
            item2["input_ids"]=tokenized_instruction["input_ids"] + label_instruction["input_ids"][1:] #because labels first token is <\s>
            item2["attention_mask"] = tokenized_instruction["attention_mask"] + label_instruction["attention_mask"][1:]
            item2["labels"]=tokenized_instruction["labels"] + label_instruction["input_ids"][1:]
        elif self.state=="test":
            item2["labels"] = tokenized_instruction["labels"] + label_instruction["input_ids"][1:]
            item2["input_ids"]=tokenized_instruction["input_ids"] + label_instruction["labels"][1:]
            item2["attenion_mask"] = tokenized_instruction["attention_mask"] + label_instruction["attention_mask"][1:]
        return item2




        item["labels"][i:]=label_instruction["input_ids"][1:]
        #item.pop("token_type_ids")
        #print(item["labels"])
        #we now need to pad to 2048 


        #print("Len labels: ", len(item["labels"]))
        #print("Len input ids: ", len(item["input_ids"]))
        #print("\n\n")
        #try:		
        #assert len(item["labels"]) == len(item["input_ids"])
        #except AssertionError:
        #print(len(item["labels"]))
        #print(len(item["input_ids"]))
        #print(len(tokenized_instruction["input_ids"]))
        #print("\n\n")
        #break
        return item

    def tokenize(self, prompt):
        #print(prompt + labels)
        if self.state=="train":
            result_prompt=self.tokenizer(prompt, 
                    truncation=True, 
                    max_length=512,
                    padding=False,
                    return_tensors=None)
        if self.state=="test":
            result_prompt=self.tokenizer(prompt, 
                    truncation=True, 
                    max_length=512,
                    padding=True,
                    return_tensors=None)


        #print(type(result_prompt))
        #print(len(result_prompt["input_ids"]))

        #result_labels=self.tokenizer(labels,
        #		truncation=True,
        #max_length=1024,
        #		padding=False,
        #		return_tensors=None)
        #old_labels=result_labels["input_ids"].copy()
        #result_prompt["labels"]=[-100 for i in result_prompt["input_ids"]] + result_labels["input_ids"]
        #result_prompt["input_ids"]=result_prompt["input_ids"] + old_labels
        #print(result_prompt["input_ids"]
        #result_prompt["labels"] = [-100 for i in result_prompt["input_ids"]] + result_labels["input_ids"]
        #print(len(result_prompt["labels"]))
        #assert len(result_prompt["input_ids"]) == len(result_prompt["labels"])
        result_prompt["labels"]=[-100]*len(result_prompt["input_ids"])
        #print(result_prompt["labels"])	
        return result_prompt

    def __len__(self):
        return len(self.instances)

def formatting_prompts_func(examples):
    output_texts=[]
    for i in range(len(examples)):
        #print("element: ", examples[i])
        prompt=examples["prompt"][i]
        diagnoses=examples["label"][i]
        text=f"""Diagnose the patient based on the following information.
                ### Patient:
                {prompt}

                ### Diagnoses:
                {diagnoses}
                """
        output_texts.append(text)
    return output_texts

def preprocess(tokenizer, filename):
    dp=[]
    golds_input_ids=[]
    golds_attention_masks=[]
    prompts_input_ids=[]
    prompts_attention_masks=[]
    with open(filename, "rb") as f:
        for el in pkl.load(f):
            #print(el["prompt"])
            #print(el["label"])
            input_tokenized=tokenizer(el["prompt"], padding="max_length", max_length=1024, truncation=True)

            output_tokenized=tokenizer(el["label"], padding="max_length", max_length=1024, truncation=True)

            #print(input_tokenized)
            prompts_input_ids.append(input_tokenized["input_ids"])
            prompts_attention_masks.append(input_tokenized.attention_mask)

            golds_input_ids.append(output_tokenized.input_ids)
            golds_attention_masks.append(output_tokenized.attention_mask)
    #print(prompts_input_ids)
    #print(gold_input_ids)
    #print(prompts_attention_masks)
    #print(type(prompts_input_ids))
    assert len(prompts_input_ids) == len(prompts_attention_masks) == len(golds_input_ids) == len(golds_attention_masks)
    #test=torch.tensor(prompts_input_ids, dtype=torch.long).to("cuda")

    dataset_dict={}
    dataset_dict["input_ids"]=torch.tensor(prompts_input_ids, dtype=torch.long)
    dataset_dict["attention_mask"]=torch.tensor(prompts_attention_masks, dtype=torch.long)
    dataset_dict["labels"]=torch.tensor(golds_input_ids, dtype=torch.long)
    dataset=Dataset.from_dict(dataset_dict)
    #dataset=TensorDataset(torch.tensor(prompts_input_ids, dtype=torch.long).to("cuda"),
    #					  torch.tensor(prompts_attention_masks, dtype=torch.long).to("cuda"),
    #					  torch.tensor(golds_input_ids, dtype=torch.long).to("cuda"))
                                                  #torch.tensor(golds_attention_masks, dtype=torch.long).to("cuda"))
    print("created dataset")
    logging.info("Created dataset!")
    return dataset

def set_seed(seed:int)->None:
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic=True
    torch.backends.cudnn.benchmark=False
    os.environ["PYTHONHASHSEED"]=str(seed)
    print(f"Random seed set as: {seed}")



def train(model, train_dataset, seed, batch_size, num_epochs, learning_rate, gradient_accumulation_steps=1, model_save_path=None):
    print("Started training")
    model.train()
    set_seed(seed)
    optimizer=AdamW(model.parameters(), lr=learning_rate, eps=1e-8, weight_decay=0.1)
    train_sampler = RandomSampler(train_dataset)
    train_dataloader=DataLoader(train_dataset, sampler=train_sampler, batch_size=batch_size)
    num_training_steps=num_epochs*len(train_dataloader)
    lr_scheduler=get_scheduler(name="linear", optimizer=optimizer, num_warmup_steps=10, num_training_steps=num_training_steps)
    #optimizer=AdamW(model.parameters(), lr=learning_rate, eps=1e-8, weight_decay=0.1)

    #model, optimizer, train_dataloader, lr_scheduler=accelerator.prepare(model, optimizer, train_dataloader, lr_scheduler)
    #loss_fct=CrossEntropyLoss(ignore_index=-100)
    model.zero_grad()
    for epoch in tqdm(range(num_epochs)):
        for index, batch in enumerate(train_dataloader):
            #optimizer.zero_grad()
            #model.zero_grad()
            input_ids=batch[0]
            attention_mask=batch[1]
            labels=batch[2]
            outputs=model(input_ids=input_ids, attention_mask=attention_mask, labels=labels)
            #print(outputs.loss)

            #print(loss)
            #logging.info(loss)
            #print(outputs.logits.size())
            #aggregated_outputs=torch.cat(outputs.logits, dim=0)
            #print(aggregated_outputs)
            #loss=loss_fct(outputs.logits.view(-1, outputs.logits.size(-1)), labels.view(-1))
            #print(loss) 
            loss=outputs.loss/gradient_accumulation_steps
            accelerator.backward(loss)
            print(loss)
            #logging.info(loss)
            #accelerator.backward(loss)
            #if gradient_accumulation_steps != -1:
                #loss=loss_fct(outputs.logits.view(-1, outputs.logits.size(-1)), labels.view(-1))
                #loss=outputs.loss/gradient_accumulation_steps
                #loss.backward()
                #print("Loss: ", loss) 
                #logging.info(loss)
                #loss.backward()
            if ((index + 1) % gradient_accumulation_steps == 0) or (index+1==len(train_dataloader)):
                #print("Update")
                #accelerator.backward(loss)
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()
            #else:
                #	loss=loss_fct(outputs.logits.view(-1, outputs.logits.size(-1)), labels.view(-1))
                #	print(loss) 
                #	logging.info(loss)
                #	loss.backward()
                #	optimizer.step()
                #	lr_scheduler.step()
            model.zero_grad()

            #model.zero_grad()
    #if model_save_path != None:
     #   torch.save(model, model_save_path)
    
    
    model.save_pretrained(model_save_path, is_main_process=accelerator.is_main_process, save_function=accelerator.save, state_dict=accelerator.get_state_dict(model))
    #accelerator.save_state(accelerator.unwrap_model(model).state_dict(), model_save_path)
    #success=model.save_checkpoint(model_save_path)
    #unwrapped_model=accelerator.unwrap_model(model)
    #unwrapped_model.save_checkpoint(model_save_path)
    #fp32_model=load_state_dict_from_zero_checkpoint(unwrapped_model, model_save_path)
    #state_dict=get_fp32_state_dict_from_zero_checkpoint(model_save_path)
    return model

"""
def train(model, tokenizer, prompts, golds):
    task_prefix="Generate a discharge summary: "
        inputs=[p for p in prompts]
        #print(type(inputs[0]))
        encoding=tokenizer([task_prefix+sequence for sequence in inputs],
                padding="longest",
                max_length=512,
                truncation=True,
                return_tensors="pt")
        input_ids, attention_masks=encoding.input_ids, encoding.attention_mask

        for el in golds:
            print(len(el))
        targets=[g[0] for g in golds]

        target_encoding=tokenizer(targets, padding="longest", max_length=128, truncation=True, return_tensors="pt")
        labels=target_encoding.input_ids
        labels[labels==tokenizer.pad_token_id] = -100
        loss=model(input_ids=input_ids[:9], attention_mask=attention_masks[:9], labels=labels[:9]).loss
        print(loss.item())


        #print(input_ids[10])
        #last_input=inputs[10]
        #input_ids=tokenizer(last_input, truncation=True)
        #print(inputs[10])
        test=tokenizer(inputs[10],
                max_length=512,
                truncation=True,
                return_tensors="pt")
        #print(len(input_ids[10]))
        output=model.generate(test.input_ids)
        print(output)
        print(tokenizer.decode(output[0], skip_special_tokens=True))

"""

if __name__=="__main__":
    parser=argparse.ArgumentParser()

    parser.add_argument(
            "--model",
            "-m",
            help="Type of model to load"
            )
    parser.add_argument(
                "--tokenizer",
                "-t",
                help="Name of Tokenizer to load from huggingface"
                )
    parser.add_argument(
                "--dataset",
                help="Path to train dataset"
                )
    parser.add_argument(
                "--seed",
                type=int,
                help="Random seed for training"
                )
    parser.add_argument(
                "--batch_size",
                "-b",
                type=int,
                help="Batch size for training"
                )
    parser.add_argument(
                "--epochs",
                "-e",
                type=int,
                help="Number of epochs for training"
                )
    parser.add_argument(
                "--learning_rate",
                "-lr",
                type=float,
                help="Learning rate for training"
                )
    parser.add_argument(
                "--model_save_path",
                "-msp",
                help="Path to save model"
                )
    parser.add_argument(
                "--do_eval",
                action="store_true",
                help="Whether to immediately evaluate the model after training"
                )

    parser.add_argument(
                "--test_dataset",
                "-td",
                help="Path to test dataset"
                )
    parser.add_argument(
                "--test_batch_size",
                "-tbs",
                type=int,
                help="Batch size for testing"
                )
    parser.add_argument(
                "--log",
                "-l",
                help="log for testing",
                )
    parser.add_argument(
                "--topk",
                type=float,
                help="Top k value for generation"
                )
    parser.add_argument(
                "--temp",
                type=float,
                help="Temperature value for generation"
                )
    parser.add_argument(
                "--num_beams",
                type=int,
                help="Number of beams"
                )
    parser.add_argument(
                "--early_stopping",
                action="store_true",
                help="Apply early stopping during generatioon?"
                )
    parser.add_argument(
                "--no_rep_ngram",
                type=int,
                help="How many ngrams should not be repeated"
                )
    parser.add_argument(
                "--num_return_sequences",
                type=int,
                help="Number of return sequences"
                )
    parser.add_argument(
                "--do_sample",
                action="store_true",
                help="Whether to sample"
                )
    parser.add_argument(
                "--metrics",
                nargs="+",
                help="metrics for eval"
                )
    parser.add_argument(
                "--ff",
                action="store_true",
                help="From flax Flag for clinical t5")

    parser.add_argument(
                "--rank",
                type=int,
                help="Rank")

    parser.add_argument(
                "--world_size",
                type=int,
                help="WS")
    parser.add_argument(
                "--gradient_accumulation_steps",
                type=int,
                help="Gradient accumulation steps"
                )


    args=parser.parse_args()
    run()
    #main()

    #all_data, prompts, golds=preprocess("./dataset.pkl")
    #tokenizer=T5Tokenizer.from_pretrained("t5-small")
    #tokenizer=AutoTokenizer.from_pretrained("luqh/ClinicalT5-large")
    #tokenizer.encode("F<G>48.59178082191781<A>white<E>Secondary malig neo NEC | Second malig neo pleura | Hx of breast malignancy |<D>Cholecystectomy | Intraoper cholangiogram |<P<M>swab : nan |<T>tidal volume 33.6 pg | platelet count 99.0 fl | specimen type 13.6 seconds | calculated total co2 19.1 % | intubated 3.78 m/ul | anisocytosis 99.0 meq/l | bands 124.0 mg/dl |<L>")
    #model=T5ForConditionalGeneration.from_pretrained("t5-small")
    #model=T5ForConditionalGeneration.from_pretrained("luqh/ClinicalT5-large")
    #model=AutoModelForCausalLM.from_pretrained("microsoft/biogpt")
    #dataset=preprocess(tokenizer, "./train/train_dataset_diagnoses_icd_48.pkl")
    #train(model,dataset, 42, 8, 3, 0.00005, model_save_path="t5_small_10000_icd_48.pt")
    #train(model, tokenizer, prompts, golds)


