"""T5 Finetuning Model for discharge summary"""
from tqdm import tqdm
from transformers import T5ForConditionalGeneration, T5Tokenizer, AutoTokenizer, AutoModelForCausalLM #, BioGptForCausalLM, AutoConfig, T5Config
import torch
import pickle as pkl
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler, TensorDataset
import logging
import os
import numpy as np
import random
from torch import nn
from transformers import get_scheduler, AdamW
from torch.nn.parallel import DataParallel 
import accelerate
#from evaluate_model import evaluate_model_loop
#import torch.distributed as dist
#import torch.multiprocessing as mp
import argparse
#from peft import get_peft_config, get_peft_model, get_peft_model_state_dict, LoraConfig, TaskType
#from evaluate_model import evaluate_model_loop
#from torch.nn import CrossEntropyLoss
#from accelerate import infer_auto_device_map, Accelerator, init_empty_weights
#logging.basicConfig(filename='log_t5.log',  level=logging.DEBUG)
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
#available_gpus=torch.cuda.device_count()
#accelerator = Accelerator()
#device =accelerator.device
#print(device.index)
#offload="./" 

def run():
	logging.basicConfig(filename=args.log, level=logging.DEBUG)
	logging.info("Starting Training")
	print("Started Training")
	
	#config=AutoConfig.from_pretrained("facebook/opt-13b")
	#with init_empty_weights():
	#	model=AutoModelForCausalLM.from_config(config)
	#device_map=infer_auto_device_map(model, no_split_module_classes=["OPTDecoderLayer"])
	#device_map["lm_head"]=device_map["decoder.embed_tokens"]
	
	#print(device_map)
	#model=AutoModelForCausalLM.from_pretrained("facebook/opt-13b", from_flax=args.ff, device_map="auto")
	model=T5ForConditionalGeneration.from_pretrained(args.model, from_flax=args.ff, output_attentions=True) #, device_map="auto")
	#peft_config=LoraConfig(task_type=TaskType.SEQ_2_SEQ_LM, inference_mode=False, r=8, lora_alpha=32, lora_dropout=0.01)
	#model=get_peft_model(model, peft_config).cuda()
	#model.print_trainable_parameters()
	#model=DataParallel(model)
	

	#model=accelerator.prepare(model)
	#print(model)


	#model=BioGptForCausalLM.from_pretrained(args.model, from_flax=args.ff, device_map="auto")
	#torch.distributed.init_process_group(backend="nccl", rank=rank, world_size=world_size)
	#dist.get_rank()
	#print("Rank: ", dist.get_rank())
	##device_id=rank% torch.cuda.device_count()
	#print(device_id)
	#model = torch.nn.DataParallel(model)
	tokenizer=AutoTokenizer.from_pretrained(args.tokenizer)
	
	print("Model loaded")
	logging.info("Loading dataset")
	print(args.dataset)
	dataset=preprocess(tokenizer, args.dataset)
	test_dataset=preprocess(tokenizer, args.test_dataset)


	logging.info("Run Train loop")
	model_updated=train(model, dataset, args.seed, args.batch_size, args.epochs, args.learning_rate,args.config_name, test_dataset, args.test_batch_size, tokenizer, args.top_k, args.temp, args.num_beams, args.early_stopping, agrs.no_rep_ngram,  args.num_return_sequences, args.metrics, args.do_sample, args.gradient_accumulation_steps, args.model_save_path)
	logging.info("Training Done!")
	if args.do_eval == True:
		logging.info("Running Evaluation")
		logging.info("Preparing Dataset")
		test_dataset=preprocess(tokenizer, args.test_dataset)
		print(args.metrics)
		res=evaluate_model_loop(model_updated, test_dataset, args.test_batch_size, tokenizer, args.topk, args.temp, args.num_beams, args.early_stopping, args.no_rep_ngram, args.num_return_sequences, metrics=args.metrics, generative=True)
		logging.info(res)



class DiagnosesDataset(torch.utils.data.Dataset):
	def __init__(self, instances, tokenizer):
		self.instances=instances
		#self.labels=labels
		self.tokenizer=tokenizer
		
	def __getitem__(self, idx):
		item={}
		#print(self.instances[0])
		instance=self.instances[idx]
		#print(instance)
		prompt= instance["prompt"]
		labels =instance["label"]
		
		#item=self.tokenize(prompt+labels)
		tokenized_instruction=self.tokenize(prompt)
		label_instruction=self.tokenizer(labels)
		item["input_ids"]=tokenized_instruction["input_ids"]
		item["attention_mask"]=tokenized_instruction["attention_mask"]
		item["label"]=torch.tensor(label_instruction["input_ids"], dtyp=torch.float32)
		#i=len(tokenized_instruction["input_ids"])
		#while 1<len(item["input_ids"])
		#print("Len of item labels before ", len(item["labels"]))
		#item["labels"][i:]=label_instruction["input_ids"]
		#item.pop("token_type_ids")
		#print(item["labels"])
		#we now need to pad to 2048 
		#target_ids=self.tokenize(

		#print("Len labels: ", len(item["labels"]))
		#print("Len input ids: ", len(item["input_ids"]))
		#print("\n\n")
		#try:		
		#	assert len(item["labels"]) == len(item["input_ids"])
		#except AssertionError:
		#	print(len(item["labels"]))
		#	print(len(item["input_ids"]))
		#	print(len(tokenized_instruction["input_ids"]))
		#	print("\n\n")
		#	break
		return item

	def tokenize(self, prompt):
		#print(prompt + labels)
		result_prompt=self.tokenizer(prompt, 
				truncation=True, 
				max_length=512,
				padding="max_length",
				return_tensors="pt")
		
		#print(type(result_prompt))
		#print(len(result_prompt["input_ids"]))

		#result_labels=self.tokenizer(labels,
		#		truncation=True,
		#		max_length=1024,
		#		padding=False,
		#		return_tensors=None)
		#old_labels=result_labels["input_ids"].copy()
		#result_prompt["labels"]=[-100 for i in result_prompt["input_ids"]] + result_labels["input_ids"]
		#result_prompt["input_ids"]=result_prompt["input_ids"] + old_labels
		#print(result_prompt["input_ids"]
		#result_prompt["labels"] = [-100 for i in result_prompt["input_ids"]] + result_labels["input_ids"]
		#print(len(result_prompt["labels"]))
		#assert len(result_prompt["input_ids"]) == len(result_prompt["labels"])
		result_prompt["labels"]=[-100]*len(result_prompt["input_ids"])
		#print(result_prompt["labels"])	
		return result_prompt

	def __len__(self):
		return len(self.instances)



def set_seed(seed:int)->None:
	np.random.seed(seed)
	random.seed(seed)
	torch.manual_seed(seed)
	torch.cuda.manual_seed(seed)
	torch.backends.cudnn.deterministic=True
	torch.backends.cudnn.benchmark=False
	os.environ["PYTHONHASHSEED"] = str(seed)
	print(f"Random seed set as: {seed}")
def preprocess(tokenizer, filename):
	dp=[]
	golds_input_ids=[]
	golds_attention_masks=[]
	prompts_input_ids=[]
	prompts_attention_masks=[]
	with open(filename, "rb") as f:
		for el in pkl.load(f):
			#print(el["prompt"])
			#print(el["label"])
			input_tokenized=tokenizer(el["prompt"], padding="max_length", max_length=512, truncation=True)
			
			output_tokenized=tokenizer(el["label"], padding="max_length", max_length=512, truncation=True)
			
			#print(input_tokenized)
			prompts_input_ids.append(input_tokenized["input_ids"])
			prompts_attention_masks.append(input_tokenized.attention_mask)
			
			golds_input_ids.append(output_tokenized.input_ids)
			golds_attention_masks.append(output_tokenized.attention_mask)
	#print(prompts_input_ids)
	#print(gold_input_ids)
	#print(prompts_attention_masks)
	#print(type(prompts_input_ids))
	assert len(prompts_input_ids) == len(prompts_attention_masks) == len(golds_input_ids) == len(golds_attention_masks)
	#test=torch.tensor(prompts_input_ids, dtype=torch.long).to("cuda")
	dataset=TensorDataset(torch.tensor(prompts_input_ids, dtype=torch.long).to("cuda"),
						  torch.tensor(prompts_attention_masks, dtype=torch.long).to("cuda"),
						  torch.tensor(golds_input_ids, dtype=torch.long).to("cuda"))
						  #torch.tensor(golds_attention_masks, dtype=torch.long).to("cuda"))
	print("created dataset")
	logging.info("Created dataset!")
	return dataset

def train(model, train_dataset, seed, batch_size, num_epochs, learning_rate, config_name, test_dataset, test_batch_size, tokenizer, top_k, temp, num_beams, early_stopping, no_rep, num_return_sequences, metrics, do_sample, gradient_accumulation_steps=1, model_save_path=None):
	print("Started training")
	model.train().to(device)
	set_seed(seed)
	optimizer=AdamW(model.parameters(), lr=learning_rate, eps=1e-8, weight_decay=0.1)
	train_sampler = RandomSampler(train_dataset)
	train_dataloader=DataLoader(train_dataset, sampler=train_sampler, batch_size=batch_size)
	num_training_steps=num_epochs*len(train_dataloader)
	lr_scheduler=get_scheduler(name="linear", optimizer=optimizer, num_warmup_steps=10, num_training_steps=num_training_steps)
	#optimizer=AdamW(model.parameters(), lr=learning_rate, eps=1e-8, weight_decay=0.1)
	
	#model, optimizer, train_dataloader, lr_scheduler=accelerator.prepare(model, optimizer, train_dataloader, lr_scheduler)
	#loss_fct=CrossEntropyLoss(ignore_index=-100)
	model.zero_grad()
	for epoch in tqdm(range(num_epochs)):
		for index, batch in enumerate(train_dataloader):
			input_ids=batch[0]
			attention_mask=batch[1]
			labels_ids=batch[2]
			labels_ids[labels_ids[:, :] == 0] = -100
			
			outputs=model(input_ids=input_ids, attention_mask=attention_mask, labels=labels_ids)
			 
			loss=outputs.loss/gradient_accumulation_steps
			loss.backward()
			print(loss)
			if ((index + 1) % gradient_accumulation_steps == 0) or (index+1==len(train_dataloader)):
				print("Update")
				#accelerator.backward(loss)
				optimizer.step()
				lr_scheduler.step()
				optimizer.zero_grad()
			

			model.zero_grad()
		epoch_prec, epoch_rec, epoch_f1=evaluate_model.evaluate_model_loop(model, config_name, test_dataset, test_batch_size, tokenizer, top_k, temp, num_beams, early_stopping, no_rep, num_return_sequences, metrics, do_sample)
	#if model_save_path != None:
	#	torch.save(model, model_save_path)
		model.save_pretrained(model_save_path/model_save_path + str(epoch))
	

	return model

"""
def train(model, tokenizer, prompts, golds):
	task_prefix="Generate a discharge summary: "
	inputs=[p for p in prompts]
	#print(type(inputs[0]))
	encoding=tokenizer([task_prefix+sequence for sequence in inputs],
		padding="longest",
		max_length=512,
		truncation=True,
		return_tensors="pt")
	input_ids, attention_masks=encoding.input_ids, encoding.attention_mask

	for el in golds:
		print(len(el))
	targets=[g[0] for g in golds]
	
	target_encoding=tokenizer(targets, padding="longest", max_length=128, truncation=True, return_tensors="pt")
	labels=target_encoding.input_ids
	labels[labels==tokenizer.pad_token_id] = -100
	loss=model(input_ids=input_ids[:9], attention_mask=attention_masks[:9], labels=labels[:9]).loss
	print(loss.item())

	
	#print(input_ids[10])
	#last_input=inputs[10]
	#input_ids=tokenizer(last_input, truncation=True)
	#print(inputs[10])
	test=tokenizer(inputs[10],
		max_length=512,
		truncation=True,
		return_tensors="pt")
	#print(len(input_ids[10]))
	output=model.generate(test.input_ids)
	print(output)
	print(tokenizer.decode(output[0], skip_special_tokens=True))

"""

if __name__=="__main__":
	parser=argparse.ArgumentParser()

	parser.add_argument(
		"--model",
		"-m",
		help="Type of model to load"
		)
	parser.add_argument(
		"--tokenizer",
		"-t",
		help="Name of Tokenizer to load from huggingface"
		)
	parser.add_argument(
		"--dataset",
		help="Path to train dataset"
		)
	parser.add_argument(
		"--seed",
		type=int,
		help="Random seed for training"
		)
	parser.add_argument(
		"--batch_size",
		"-b",
		type=int,
		help="Batch size for training"
		)
	parser.add_argument(
		"--epochs",
		"-e",
		type=int,
		help="Number of epochs for training"
		)
	parser.add_argument(
		"--learning_rate",
		"-lr",
		type=float,
		help="Learning rate for training"
		)
	parser.add_argument(
		"--model_save_path",
		"-msp",
		help="Path to save model"
		)
	parser.add_argument(
		"--do_eval",
		action="store_true",
		help="Whether to immediately evaluate the model after training"
		)
	
	parser.add_argument(
		"--test_dataset",
		"-td",
		help="Path to test dataset"
		)
	parser.add_argument(
		"--test_batch_size",
		"-tbs",
		type=int,
		help="Batch size for testing"
		)
	parser.add_argument(
		"--log",
		"-l",
		help="log for testing",
		)
	parser.add_argument(
		"--top_k",
		type=float,
		help="Top k value for generation"
		)
	parser.add_argument(
		"--temp",
		type=float,
		help="Temperature value for generation"
		)
	parser.add_argument(
		"--num_beams",
		type=int,
		help="Number of beams"
		)
	parser.add_argument(
		"--early_stopping",
		action="store_true",
		help="Apply early stopping during generatioon?"
		)
	parser.add_argument(
		"--no_rep_ngram",
		type=int,
		help="How many ngrams should not be repeated"
		)
	parser.add_argument(
		"--num_return_sequences",
		type=int,
		help="Number of return sequences"
		)
	parser.add_argument(
		"--do_sample",
		action="store_true",
		help="Whether to sample"
		)
	parser.add_argument(
		"--metrics",
		nargs="+",
		help="metrics for eval"
		)
	parser.add_argument(
		"--ff",
		action="store_true",
		help="From flax Flag for clinical t5")

	parser.add_argument(
		"--rank",
		type=int,
		help="Rank")

	parser.add_argument(
		"--world_size",
		type=int,
		help="WS")
	parser.add_argument(
		"--gradient_accumulation_steps",
		type=int,
		help="Gradient accumulation steps"
		)
		
	
	args=parser.parse_args()
	run()

	#main()

	#all_data, prompts, golds=preprocess("./dataset.pkl")
	#tokenizer=T5Tokenizer.from_pretrained("t5-small")
	#tokenizer=AutoTokenizer.from_pretrained("luqh/ClinicalT5-large")
	#tokenizer.encode("F<G>48.59178082191781<A>white<E>Secondary malig neo NEC | Second malig neo pleura | Hx of breast malignancy |<D>Cholecystectomy | Intraoper cholangiogram |<P<M>swab : nan |<T>tidal volume 33.6 pg | platelet count 99.0 fl | specimen type 13.6 seconds | calculated total co2 19.1 % | intubated 3.78 m/ul | anisocytosis 99.0 meq/l | bands 124.0 mg/dl |<L>")
	#model=T5ForConditionalGeneration.from_pretrained("t5-small")
	#model=T5ForConditionalGeneration.from_pretrained("luqh/ClinicalT5-large")
	#model=AutoModelForCausalLM.from_pretrained("microsoft/biogpt")
	#dataset=preprocess(tokenizer, "./train/train_dataset_diagnoses_icd_48.pkl")
	#train(model,dataset, 42, 8, 3, 0.00005, model_save_path="t5_small_10000_icd_48.pt")
	#train(model, tokenizer, prompts, golds)

 
